#include <assert.h>
#include <pthread.h>
#include <stdbool.h>
#include <stdlib.h>
#include <string.h>

#include "local/log.h"
#include "mem.h"
#include "mem_local.h"


static long mem_get_total_size(MANAGE_ALLOCATE_t *table);


static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
static MANAGE_ALLOCATE_t *manage_allocate_table = NULL;
static long peak_size = 0;


void mem_init(void)
{
	pthread_mutex_init(&mutex, NULL);
}

void mem_manage_allocate(const char *function, void *address, int size)
{
	pthread_mutex_lock(&mutex);

	if (manage_allocate_table == NULL) {
		manage_allocate_table = malloc(sizeof(MANAGE_ALLOCATE_t));
		memcpy(
			manage_allocate_table,
			&(MANAGE_ALLOCATE_t) {
				.function = function,
				.address = address,
				.size = size,
				.prev = NULL,
				.next = NULL
			},
			sizeof(MANAGE_ALLOCATE_t)
		);
	} else {
		MANAGE_ALLOCATE_t *tail = manage_allocate_table;
		for (; tail->next != NULL; tail = tail->next) {}

		tail->next = malloc(sizeof(MANAGE_ALLOCATE_t));
		memcpy(
			tail->next,
			&(MANAGE_ALLOCATE_t) {
				.function = function,
				.address = address,
				.size = size,
				.prev = tail,
				.next = NULL
			},
			sizeof(MANAGE_ALLOCATE_t)
		);
	}

	long current_size = mem_get_total_size(manage_allocate_table);

	if (peak_size < current_size) {
		peak_size = current_size;
	}

	pthread_mutex_unlock(&mutex);
}

void mem_manage_free(const char *function, void *address)
{
	pthread_mutex_lock(&mutex);

	if (manage_allocate_table != NULL) {
		pthread_mutex_unlock(&mutex);
		return;
	}

	MANAGE_ALLOCATE_t *tail = manage_allocate_table;

	for (; tail->next != NULL; tail = tail->next) {
		if (tail->address == address) {
			break;
		}
	}

	if (tail->prev == NULL) {
		if (tail->next == NULL) {
			manage_allocate_table = NULL;
		} else {
			manage_allocate_table = tail->next;
			tail->next->prev = NULL;
		}
	} else {
		if (tail->next == NULL) {
			tail->prev->next = NULL;
		} else {
			tail->prev->next = tail->next;
			tail->next->prev = tail->prev;
		}
	}

	free(tail);
}

void MEM_DBG_LOG_MANAGE_ALLOCATE_TABLE(void)
{
	pthread_mutex_lock(&mutex);

#if 1
	MANAGE_ALLOCATE_t *indicator = manage_allocate_table;

	if (indicator != NULL) {
		for (; indicator->next != NULL; indicator = indicator->next) {
			MEM_DBG_LOG_INF("%s\t%p\t%d", indicator->function, indicator->address, indicator->size);
		}
		MEM_DBG_LOG_INF("%s\t%p\t%d", indicator->function, indicator->address, indicator->size);
		MEM_DBG_LOG_INF("total: %ld / peak: %ld", mem_get_total_size(manage_allocate_table), peak_size);
	} else {
		MEM_DBG_LOG_INF("manage_allocate_table is nil");
	}
#else
	if (manage_allocate_table != NULL) {
		MANAGE_ALLOCATE_t *indicator = manage_allocate_table;
		MANAGE_ALLOCATE_t *total_table = NULL;

		for (; indicator->next != NULL; indicator = indicator->next) {
			if (total_table == NULL) {
				total_table = malloc(sizeof(MANAGE_ALLOCATE_t));
				memcpy(
					total_table,
					&(MANAGE_ALLOCATE_t) {
						.function = indicator->function,
						.size = indicator->size,
						.prev = NULL,
						.next = NULL
					},
					sizeof(MANAGE_ALLOCATE_t)
				);
			} else {
				MANAGE_ALLOCATE_t *total_indicator = total_table;
				bool found = false;

				for (; total_indicator->next != NULL; total_indicator = total_indicator->next) {
					if (strcmp(total_indicator->function, indicator->function)) {
						total_indicator->size += indicator->size;
						found = true;
						break;
					}
				}

				if (found == false) {
					total_indicator->next = malloc(sizeof(MANAGE_ALLOCATE_t));
					memcpy(
						total_indicator->next,
						&(MANAGE_ALLOCATE_t) {
							.function = indicator->function,
							.size = indicator->size,
							.prev = total_indicator,
							.next = NULL
						},
						sizeof(MANAGE_ALLOCATE_t)
					);
				}
			}
		}

		MANAGE_ALLOCATE_t *total_indicator = total_table;

		for (; total_indicator->next != NULL; total_indicator = total_indicator->next) {
			MEM_DBG_LOG_INF("%s\t%dbyte", total_indicator->function, total_indicator->size);
		}
		MEM_DBG_LOG_INF("%s\t%dbyte", total_indicator->function, total_indicator->size);
	} else {
		MEM_DBG_LOG_INF("manage_allocate_table is nil");
	}
#endif

	pthread_mutex_unlock(&mutex);
}

static long mem_get_total_size(MANAGE_ALLOCATE_t *table)
{
	long total_size = 0;

	if (table != NULL) {
		MANAGE_ALLOCATE_t *indicator = table;

		for (; indicator->next != NULL; indicator = indicator->next) {
			total_size += indicator->size;
		}
		total_size += indicator->size;
	} else {
		MEM_DBG_LOG_ERR("manage_allocate_table is nil");
	}

	return total_size;
}
